/* wrap.m     Simple custom wrappers for Portland example
 *
 * Copyright (c) 2011 Psellos   http://psellos.com
 * Licensed under the MIT license:
 *     http://www.opensource.org/licenses/mit-license.php
 */
# import <stdlib.h>
# import <objc/runtime.h>
# import <UIKit/UIKit.h>
# import <caml/mlvalues.h>
# import <caml/memory.h>
# import <caml/alloc.h>
# import <caml/callback.h>
# import "wrap.h"

/* Represent raw ObjC object pointers in OCaml as nativeint values.
 *
 * Note: to keep things simple, we retain ObjC values and don't release
 * them.  It works for this example because the few ObjC objects
 * accessible from the OCaml side are never destroyed.
 * 
 * In a more dynamic setting, need to gain control when the OCaml
 * objects are GCed and release the corresponding ObjC objects.
 */
static value wrapRawObjC(id obj) {
    CAMLparam0();
    [obj retain];
    CAMLreturn(caml_copy_nativeint((intnat) obj));
}

static id unwrapRawObjC(value objcval) {
    return (id) Nativeint_val(objcval);
}

/* Need to be able to determine the OCaml object that wraps a given ObjC
 * object.  In the simple setup here, the association is permanent and
 * there are very few objects involved.  So we can use a little table.
 * (Getting the ObjC object that wraps an OCaml object is easier; we
 * just use the container method of Wrappee.t.)
 */
static struct {
    id objc_obj;
    value ocaml_obj;
} g_ocaml_wrapper[16];
static int g_ocwrap_ct = 0;


static value find_ocaml_wrapper(id obj) {
    int i;

    for(i = 0; i < g_ocwrap_ct; i++)
        if(g_ocaml_wrapper[i].objc_obj == obj)
            return g_ocaml_wrapper[i].ocaml_obj;
    return 0;
}


static value wrapObjC(char *wrapcs, id obj) {
    /* Create a new OCaml object that wraps the given ObjC object.
     * wrapcs gives the name of the closure for creating an OCaml
     * wrapper of the correct type.  A closure is registered by each
     * OCaml class that wraps an ObjC class (i.e., each subclass of
     * Wrapper.t).
     */
    CAMLparam0();
    CAMLlocal1(val);
    value *closure;

    closure = (value *) caml_named_value(wrapcs);
    if(closure == NULL) {
        /* Can't create an instance of an unregistered class.  If this
         * happens, things will go very bad from here.
         */
        fprintf(stderr, "wrapObjC lookup failure for %s\n", wrapcs);
        fflush(stderr);
        CAMLreturn(Val_unit); /* Not type safe, should never happen */
    }
    val = caml_callback(*closure, wrapRawObjC(obj));

    /* Enter the wrapping into the table.
     */
    g_ocaml_wrapper[g_ocwrap_ct].objc_obj = obj;
    g_ocaml_wrapper[g_ocwrap_ct].ocaml_obj = val;
    caml_register_global_root(&g_ocaml_wrapper[g_ocwrap_ct].ocaml_obj);
    g_ocwrap_ct++;

    CAMLreturn(val);
}

static id unwrapObjC(value obj) {
    /* Return the ObjC object wrapped by the given OCaml object.
     */
    CAMLparam1(obj);
    CAMLlocal1(objcval);

    objcval =
        caml_callback(
            caml_get_public_method(obj, caml_hash_variant("contents")),
            obj);
    CAMLreturnT(id, unwrapRawObjC(objcval));
}


/* Subclasses of WrapOCaml wrap OCaml objects as ObjC objects.
 */
@implementation WrapOCaml {
}

@synthesize contents;

- (WrapOCaml *) init
{
    /* Create a new OCaml object that is wrapped by self.  We match the
     * OCaml class to our own class by name--simple but effective.  We
     * then call the closure for creating a wrapped object, which is
     * registered by each OCaml class that can be wrapped an ObjC class
     * (i.e., each subclass of Wrappee.t).
     */
    const char *classname;
    char closurename[128];
    value *closure;

    if((self = [super init]) != nil) {
        classname = object_getClassName(self);
        if(strlen(classname) + 8 >= sizeof closurename) {
            fprintf(stderr, "[WrapOCaml init]: class name too long: %s\n",
                    classname);
            fflush(stderr);
            return nil;
        }
        strcpy(closurename, classname);
        strcat(closurename, ".wrapped");
        closure = (value *) caml_named_value(closurename);
        if(closure == NULL) {
            fprintf(stderr, "[WrapOCaml init]: lookup failure for %s\n",
                    closurename);
            fflush(stderr);
            return nil;
        }
        contents = caml_callback(*closure, wrapRawObjC(self));
        caml_register_global_root(&contents);
    }
    return self;
}

- (void) dealloc
{
    if(contents != 0) {
        caml_remove_global_root(&contents);
        contents = 0;
    }
    [super dealloc];
}

@end

/* OCaml objects accessed from ObjC.
 */

@implementation Portlandappdeleg : WrapOCaml
{
}

@dynamic label;

- (UILabel *) label
{
    CAMLparam0();
    CAMLlocal2(selfval, labelval);

    selfval = [self contents];
    labelval =
        caml_callback(
            caml_get_public_method(selfval, caml_hash_variant("label")),
            selfval);
    CAMLreturnT(UILabel *, unwrapObjC(labelval));
}

- (void) setLabel: (UILabel *) aLabel
{
    CAMLparam0();
    CAMLlocal2(selfval, labelval);

    selfval = [self contents];
    labelval = wrapObjC("UiLabel.wrap", (id) aLabel);
    (void)
        caml_callback2(
            caml_get_public_method(selfval, caml_hash_variant("setLabel'")),
            selfval,
            labelval);
    CAMLreturn0;
}

- (void) applicationDidFinishLaunching: (UIApplication *) anApplication
{
    CAMLparam0();
    CAMLlocal2(selfval, applval);

    selfval = [self contents];
    if((applval = find_ocaml_wrapper(anApplication)) == 0)
        applval = wrapObjC("UiApplication.wrap", (id) anApplication);
    (void)
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("applicationDidFinishLaunching'")),
            selfval,
            applval);
    CAMLreturn0;
}

- (void) applicationWillResignActive: (UIApplication *) anApplication
{
    CAMLparam0();
    CAMLlocal2(selfval, applval);

    selfval = [self contents];
    if((applval = find_ocaml_wrapper(anApplication)) == 0)
        applval = wrapObjC("UiApplication.wrap", (id) anApplication);
    (void)
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("applicationWillResignActive'")),
            selfval,
            applval);
    CAMLreturn0;
}

- (void) applicationDidBecomeActive: (UIApplication *) anApplication
{
    CAMLparam0();
    CAMLlocal2(selfval, applval);

    selfval = [self contents];
    if((applval = find_ocaml_wrapper(anApplication)) == 0)
        applval = wrapObjC("UiApplication.wrap", (id) anApplication);
    (void)
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("applicationDidBecomeActive'")),
            selfval,
            applval);
    CAMLreturn0;
}

- (void) timerTick: (NSTimer *) aTimer
{
    CAMLparam0();
    CAMLlocal2(selfval, timerval);

    selfval = [self contents];
    if((timerval = find_ocaml_wrapper(aTimer)) == 0)
        /* Can't report a timer tick if the timer isn't known.
         */
        return;
    (void)
        caml_callback2(
            caml_get_public_method(selfval, caml_hash_variant("timerTick'")),
            selfval,
            timerval);
    CAMLreturn0;
}

@end


/* ObjC objects accessed from OCaml.
 */

/* NSTimer
 */

value NSTimer_scheduledTimer5(value timeInterval, value target, value selector,
                    value userInfo, value repeats)
    /* For simplicity, the target must be wrapped (by an ObjC object)
     * already.  It's OK if userInfo isn't wrapped; we treat this as
     * nil.
     */
{
    CAMLparam5(timeInterval, target, selector, userInfo, repeats);
    CAMLlocal2(targro, uiro);

    double ti = Double_val(timeInterval);
    targro = 
        caml_callback(
            caml_get_public_method(target, caml_hash_variant("container")),
            target);
    id targ = unwrapRawObjC(targro);
    if(targ == nil) {
        /* If the target wasn't already wrapped, things will go very bad
         * from here.
         */
        fprintf(stderr, "NSTimer_scheduledTimer5: unwrapped target\n");
        fflush(stderr);
        CAMLreturn(Val_unit); /* Not type safe, should never happen */
    }
    char *sels = String_val(selector);
    SEL sel = NSSelectorFromString([NSString stringWithUTF8String: sels]);
    uiro =
        caml_callback(
            caml_get_public_method(userInfo, caml_hash_variant("container")),
            userInfo);
    id ui = unwrapRawObjC(uiro);
    int rep = Bool_val(repeats);

    NSTimer *timer =
        [NSTimer scheduledTimerWithTimeInterval: ti
                 target: targ
                 selector: sel
                 userInfo: ui
                 repeats: rep ];
    CAMLreturn(wrapObjC("NsTimer.wrap", (id) timer));
}


/* UILabel
 */

value UILabel_setText_(value objcval, value textval)
{
    CAMLparam2(objcval, textval);

    UILabel *label = unwrapRawObjC(objcval);
    NSString *text = [NSString stringWithUTF8String: String_val(textval)];
    [label setText: text];
    CAMLreturn(Val_unit);
}


value UILabel_setBounds_(value objcval, value boundsval)
{
    CAMLparam2(objcval, boundsval);
    CGRect bounds;

    UILabel *label = unwrapRawObjC(objcval);
    bounds.origin.x = Double_val(Field(boundsval, 0));
    bounds.origin.y = Double_val(Field(boundsval, 1));
    bounds.size.width = Double_val(Field(boundsval, 2));
    bounds.size.height = Double_val(Field(boundsval, 3));

    [label setBounds: bounds];
    CAMLreturn(Val_unit);
}

value UILabel_setTransform_(value objcval, value transval)
{
    CAMLparam2(objcval, transval);
    CGAffineTransform trans;

    UILabel *label = unwrapRawObjC(objcval);
    trans.a = Double_val(Field(transval, 0));
    trans.b = Double_val(Field(transval, 1));
    trans.c = Double_val(Field(transval, 2));
    trans.d = Double_val(Field(transval, 3));
    trans.tx = Double_val(Field(transval, 4));
    trans.ty = Double_val(Field(transval, 5));

    [label setTransform: trans];
    CAMLreturn(Val_unit);
}


/* UIDevice
 */

value UIDevice_currentDevice(value unit) {
    CAMLparam1(unit);
    CAMLlocal1(deviceval);

    UIDevice *device = [UIDevice currentDevice];
    if((deviceval = find_ocaml_wrapper(device)) == 0)
        deviceval = wrapObjC("UiDevice.wrap", (id) device);
    CAMLreturn(deviceval);
}

value UIDevice_beginGeneratingDeviceOrientationNotifications(value objcval)
{
    CAMLparam1(objcval);

    UIDevice *device = unwrapRawObjC(objcval);
    [device beginGeneratingDeviceOrientationNotifications];
    CAMLreturn(Val_unit);
}

value UIDevice_orientation(value objcval)
{
    CAMLparam1(objcval);

    UIDevice *device = unwrapRawObjC(objcval);
    UIDeviceOrientation orient = [device orientation];
    if(     orient < UIDeviceOrientationUnknown ||
            orient > UIDeviceOrientationFaceDown)
        orient = UIDeviceOrientationUnknown;
    CAMLreturn(Val_int((int) orient));
}
